import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import uptrain.v0 as v0


# if __name__ == "__main__":
def test_a_b_testing():
    """Test A/B testing using concept drift ADWIN algorithm.

    Refer to the documentation in the test for concept drift for more information
    on the concept drift detection using ADWIN algorithm part.

    Here, we will use the same concept drift detection algorithm to detect A/B testing.
    We have two models that we want to compare:
        - both models predict the ground truths almost correctly for the first 1000 values
        - model_1 predicts the ground truths almost correctly for the next 500 values
        - model_2, however, has higher variance in its predictions for the next 500 values
    We would like to compare the performance of both models on the same data and see
    how concept drift detection is used in A/B testing here.

    In this test, we use the UpTrain framework to log the inputs, outputs, and ground truths.
    We set up the framework to use the ADWIN algorithm to monitor the concept drift of the
    different models. We use one error rate for detection of concept drift:
        - The error rate is calculated as the mean absolute difference between the
          predictions and the ground truths.

    The test passes if the ADWIN algorithm successfully detects the concept drift and raises
    an alert around the point where predictions from the models start becoming inaccurate.
    """

    random_state = np.random.RandomState(seed=1337)
    n = 1000
    params_model_1 = [(0.2, 0.03, n), (0.5, 0.08, n)]
    params_model_2 = [(0.2, 0.03, n), (0.55, 0.1, n)]
    distributions_model_1 = np.array(
        [random_state.normal(*param) for param in params_model_1]
    )
    distributions_model_2 = np.array(
        [random_state.normal(*param) for param in params_model_2]
    )
    ground_truths = np.array([0.2] * n + [0.5] * (n // 2) + [0.6] * (n // 2))
    stream_model_1 = distributions_model_1.flatten()
    stream_model_2 = distributions_model_2.flatten()

    def plot_data(stream, distributions, drifts=None):
        """Helper function to plot data."""

        _ = plt.figure(figsize=(7, 3), tight_layout=True)
        gridspecs = gridspec.GridSpec(1, 2, width_ratios=[3, 1])

        ax1, ax2 = plt.subplot(gridspecs[0]), plt.subplot(gridspecs[1])
        ax1.grid()
        ax1.plot(stream, label="stream")
        ax2.grid(axis="y")

        for index, dist in enumerate(distributions):
            ax2.hist(dist, label=f"dist_{index}")

        drift_color = "red"
        if drifts is not None:
            for drift_detected in drifts:
                ax1.axvline(drift_detected, color=drift_color)
        plt.tight_layout()
        plt.show()

    # Plot data for debugging
    # plot_data(stream_model_1, distributions_model_1)
    # plot_data(stream_model_2, distributions_model_2)

    # Create a configuration for the framework
    cfg = {
        # Define checks to be performed
        "checks": [
            # First check: monitor concept drift using ADWIN algorithm with
            # mean absolute error as the measurable for the feature "data".
            {
                "type": v0.Monitor.CONCEPT_DRIFT,
                "algorithm": v0.DataDriftAlgo.ADWIN,
                "measurable_args": {"type": v0.MeasurableType.MAE},
                # Configurable parameters that the ADWIN algorithm supports
                "delta": 0.002,
                "clock": 32,
                "max_buckets": 5,
                "min_window_length": 5,
                "grace_period": 5,
                # Add arguments to handle multiple models for A/B testing
                "model_args": [
                    {
                        "type": v0.MeasurableType.INPUT_FEATURE,
                        "feature_name": "model_name",
                        "allowed_values": ["model_1", "model_2"],
                    }
                ],
            },
        ],
        # Specify where the logging data should be stored
        "retraining_folder": "uptrain_a_b_testing",
        # Specify logging arguments
        # "st_logging" should be True if we want streamlit logging, False otherwise
        "logging_args": {"st_logging": True},
    }

    framework = v0.Framework(cfg)
    batch_size = 64
    size = len(stream_model_1)

    # Feed the data to the framework
    for i in range(size // batch_size):
        data_model_1 = stream_model_1[i * batch_size : (i + 1) * batch_size]
        data_model_2 = stream_model_2[i * batch_size : (i + 1) * batch_size]
        data = np.concatenate([data_model_1, data_model_2])
        inputs = {
            "data": data,
            "model_name": ["model_1"] * len(data_model_1)
            + ["model_2"] * len(data_model_2),
        }
        outputs = data
        gts = ground_truths[i * batch_size : (i + 1) * batch_size]
        gts = np.concatenate([gts, gts])
        ids = framework.log(inputs=inputs, outputs=outputs)
        framework.log(identifiers=ids, gts=gts)



"""
This module implements a simple LLM wrapper that can be used to evaluate performance
of LLM applications. 
"""

import typing as t

from loguru import logger
import httpx
import polars as pl
import pandas as pd
import pydantic

from uptrain.framework.remote import APIClientWithoutAuth, DataSchema
from uptrain.framework.base import Settings

from uptrain.framework.evals import Evals, ParametricEval, CritiqueTone, GuidelineAdherence, ResponseMatching, ConversationSatisfaction


class EvalLLM:

    def __init__(self, settings: Settings = None, openai_api_key: str = None) -> None:
        if (openai_api_key is None) and (settings is None):
            raise Exception("Please provide OpenAI API Key")

        if settings is None:
            self.settings = Settings(openai_api_key=openai_api_key)
        else:
            self.settings = settings
        self.executor = APIClientWithoutAuth(self.settings)

    def evaluate(
        self,
        data: t.Union[list[dict], pl.DataFrame, pd.DataFrame],
        checks: list[t.Union[str, Evals, ParametricEval]],
        schema: t.Union[DataSchema, dict[str, str], None] = None,
        metadata: t.Optional[dict[str, str]] = None
    ):
        """Run an evaluation on the UpTrain server using user's openai keys.
        NOTE: This api doesn't log any data.

        Args:
            data: Data to evaluate on. Either a Pandas DataFrame or a list of dicts.
            checks: List of checks to evaluate on.
            schema: Schema of the data. Only required if the data attributes aren't typical (question, response, context).

        Returns:
            results: List of dictionaries with each data point and corresponding evaluation results.
        """

        if isinstance(data, pl.DataFrame):
            data = data.to_dicts()

        if schema is None:
            schema = DataSchema()
        elif isinstance(schema, dict):
            schema = DataSchema(**schema)

        if metadata is None:
            metadata = {}

        checks = [Evals(m) if isinstance(m, str) else m for m in checks]
        for m in checks:
            assert isinstance(m, (Evals, ParametricEval))

        req_attrs, ser_checks = set(), []
        for m in checks:
            if m in [Evals.FACTUAL_ACCURACY, Evals.RESPONSE_COMPLETENESS_WRT_CONTEXT, Evals.RESPONSE_CONSISTENCY]:
                req_attrs.update([schema.question, schema.context, schema.response])
            elif m in [Evals.RESPONSE_RELEVANCE, Evals.RESPONSE_COMPLETENESS, Evals.RESPONSE_CONCISENESS, Evals.PROMPT_INJECTION]:
                req_attrs.update([schema.question, schema.response])
            elif m in [Evals.CONTEXT_RELEVANCE]:
                req_attrs.update([schema.question, schema.context])
            elif m in [Evals.CRITIQUE_LANGUAGE, Evals.VALID_RESPONSE] or isinstance(m, CritiqueTone) or isinstance(m, GuidelineAdherence):
                req_attrs.update([schema.response])
            elif isinstance(m, ResponseMatching):
                req_attrs.update([schema.response, schema.ground_truth])
            elif isinstance(m, ConversationSatisfaction):
                req_attrs.update([schema.conversation])

            if isinstance(m, ParametricEval):
                ser_checks.append({"check_name": m.__class__.__name__, **m.dict()})
            elif isinstance(m, Evals):
                ser_checks.append(m.value)
            else:
                raise ValueError(f"Invalid metric: {m}")
        for idx, row in enumerate(data):
            if not req_attrs.issubset(row.keys()):
                raise ValueError(
                    f"Row {idx} is missing required all required attributes for evaluation: {req_attrs}"
                )

        # send in chunks of 50, so the connection doesn't time out waiting for the server
        results = []
        NUM_TRIES, BATCH_SIZE = 3, 50
        for i in range(0, len(data), BATCH_SIZE):
            batch_results = []
            for try_num in range(NUM_TRIES):
                try:
                    logger.info(
                        f"Sending evaluation request for rows {i} to <{i+BATCH_SIZE} to the Uptrain"
                    )
                    batch_results = self.executor.evaluate(
                        data=data[i : i + BATCH_SIZE],
                        checks=ser_checks,
                        metadata={
                            "schema": schema.dict(),
                            "uptrain_settings": self.settings.dict()
                        }
                    )

                except Exception as e:
                    logger.info("Retrying evaluation request")
                    if try_num == NUM_TRIES - 1:
                        logger.error(f"Evaluation failed with error: {e}")
                        raise e

            results.extend(batch_results)

        return results



from abc import ABC

try:
    import hdbscan
except:
    hdbscan = None
from sklearn.cluster import DBSCAN
from typing import Any, Dict, Union

from uptrain.v0.constants import ClusteringAlgorithm
from uptrain.v0.core.lib.helper_funcs import dependency_required


class Clustering(ABC):
    def __init__(self):
        super().__init__()

    def resolve(self, args: Dict[str, Any]) -> Any:
        raise NotImplementedError("Clustering.resolve() not implemented.")


class DBSCANClustering(Clustering):
    def __init__(self) -> None:
        super().__init__()

    # https://scikit-learn.org/stable/modules/generated/sklearn.cluster.DBSCAN.html
    def resolve(self, args: Dict[str, Any]) -> DBSCAN:
        if not isinstance(args, dict):
            raise ValueError("args must be a dictionary.")

        valid_keys = {
            "eps",
            "min_samples",
            "metric",
            "metric_params",
            "algorithm",
            "leaf_size",
            "p",
            "n_jobs",
        }
        invalid_keys = set(args.keys()) - valid_keys

        if invalid_keys:
            raise ValueError(f"Invalid key(s) {invalid_keys} in args.")

        return DBSCAN(**args)


@dependency_required(hdbscan, "hdbscan")
class HDBSCANClustering(Clustering):
    def __init__(self) -> None:
        super().__init__()

    # https://hdbscan.readthedocs.io/en/latest/basic_hdbscan.html
    def resolve(self, args: Dict[str, Any]) -> "hdbscan.HDBSCAN":
        if not isinstance(args, dict):
            raise ValueError("args must be a dictionary.")

        valid_keys = {
            "algorithm",
            "alpha",
            "approx_min_span_tree",
            "gen_min_span_tree",
            "leaf_size",
            "memory",
            "metric",
            "min_cluster_size",
            "min_samples",
            "p",
        }
        invalid_keys = set(args.keys()) - valid_keys

        if invalid_keys:
            raise ValueError(f"Invalid key(s) {invalid_keys} in args.")

        return hdbscan.HDBSCAN(**args)


class ClusteringResolver:
    """Class that resolves a clustering key to a clustering class instance."""

    def __init__(self, algorithm: ClusteringAlgorithm) -> None:
        super().__init__()
        self.algorithm = algorithm

    def resolve(self, args: Dict[str, Any]) -> Union[DBSCAN, "hdbscan.HDBSCAN"]:
        if self.algorithm == ClusteringAlgorithm.DBSCAN:
            return DBSCANClustering().resolve(args)
        elif self.algorithm == ClusteringAlgorithm.HDBSCAN:
            return HDBSCANClustering().resolve(args)
        else:
            raise ValueError(f"Clustering algorithm {self.algorithm} not supported.")



from uptrain.operators import register_custom_op, ColumnOp


@register_custom_op
class Sum(ColumnOp):
    col_a: str
    col_b: str

    def setup(self, settings):
        return self

    def run(self, data):
        import polars as pl

        return {
            "output": data.with_columns(
                [(pl.col(self.col_a) + pl.col(self.col_b)).alias("sum")]
            )
        }


def test_custom_op():
    """Test the custom operator by serializing and deserializing it."""
    import polars as pl
    from uptrain.framework import Check, Settings

    ds = pl.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    op_sum = Sum(col_a="a", col_b="b")
    check = Check(name=op_sum, operators=[op_sum], plots=[])

    out = check.setup(Settings()).run(ds)
    assert out["sum"].series_equal(pl.Series("sum", [5, 7, 9]))

    serialized = check.dict()
    check_2 = Check.from_dict(serialized)
    out_2 = check_2.setup(Settings()).run(ds)
    assert out_2["sum"].series_equal(pl.Series("sum", [5, 7, 9]))



import os
import json
import copy
import csv
import time
from typing import Optional
from datetime import datetime, timedelta, tzinfo
from collections import OrderedDict
import functools

import numpy as np
import pandas as pd
from sklearn.cluster import KMeans
from uptrain.v0.core.encoders import UpTrainEncoder


def cluster_and_plot_data(
    data,
    num_clusters,
    cluster_plot_func=None,
    plot_save_name="",
    normalisation=None,
    compute_point_density=False,
):
    kmeans = KMeans(n_clusters=num_clusters, random_state=1, n_init=10)
    kmeans.fit(data)
    all_clusters = kmeans.cluster_centers_
    all_labels = kmeans.labels_

    counts = np.zeros(num_clusters)
    uniq_lbs, uniq_cts = np.unique(all_labels, return_counts=True)
    for idx in range(uniq_lbs.shape[0]):
        counts[uniq_lbs[idx]] = uniq_cts[idx]

    cluster_vars = []
    cluster_dictn = {}
    for idx in range(len(all_clusters)):
        this_elems = data[np.where(all_labels == idx)[0]]
        cluster_dictn.update({idx: this_elems})
        cluster_vars.append(
            np.mean(np.sum(np.abs(this_elems - all_clusters[idx]), axis=1))
        )

    density_arr = []
    if compute_point_density:
        min_var = np.min(cluster_vars)
        for idx in range(len(data)):
            cluster_centroid_dists = np.sum(np.abs(all_clusters - data[idx]), axis=1)
            closest_cluster_labels = np.where(cluster_centroid_dists < min_var * 4)
            num_close_points = 0
            for this_label in closest_cluster_labels[0]:
                this_elems = cluster_dictn[this_label]
                all_elems_dists = np.sum(np.abs(this_elems - data[idx]), axis=1)
                closest_points_idxs = np.where(all_elems_dists < 1.5 * min_var)
                num_close_points += len(closest_points_idxs[0])
            density_arr.append(num_close_points - 1)
    density_arr = np.array(density_arr)

    idxs_closest_to_cluster_centroids = {}
    for idx in range(len(all_clusters)):
        this_elems = cluster_dictn[idx]
        all_elems_dists = np.sum(np.abs(this_elems - all_clusters[idx]), axis=1)
        closest_idxs = np.argsort(all_elems_dists)[0:3]
        idxs_closest_to_cluster_centroids.update({idx: closest_idxs})

    dictn = []
    for idx in range(len(all_clusters)):
        dictn.append(
            {
                "cluster": all_clusters[idx],
                "count": counts[idx],
                "var": cluster_vars[idx],
                "idxs_closest": idxs_closest_to_cluster_centroids[idx],
            }
        )
    dictn.sort(key=lambda x: x["count"], reverse=True)

    all_clusters = np.array([x["cluster"] for x in dictn])
    counts = np.array([x["count"] for x in dictn])
    cluster_vars = np.array([x["var"] for x in dictn])
    idxs_closest_to_cluster_centroids = dict(
        zip(range(len(dictn)), [x["idxs_closest"] for x in dictn])
    )

    if normalisation is not None:
        all_clusters_renormalised = copy.deepcopy(all_clusters) * normalisation
    else:
        all_clusters_renormalised = all_clusters
    if cluster_plot_func is not None:
        cluster_plot_func(
            all_clusters_renormalised, counts, plot_save_name=plot_save_name
        )
    return (
        all_clusters,
        counts,
        cluster_vars,
        density_arr,
        idxs_closest_to_cluster_centroids,
    )


def add_data_to_warehouse(data, path_csv, row_update=False):
    """Stores the dictionary "data" at location path_csv"""

    for k in list(data.keys()):
        if row_update and k == "id":
            continue
        data[k] = [json.dumps(x, cls=UpTrainEncoder) for x in data[k]]
    if not os.path.exists(path_csv):
        pd.DataFrame(data).to_csv(path_csv, index=False)
    else:
        if row_update:
            df = pd.read_csv(path_csv)
            for k in list(data.keys()):
                if row_update and k == "id":
                    continue
                if k not in list(df.columns):
                    df[k] = None
                df.loc[get_df_indices_from_ids(df, data["id"]), k] = np.array(
                    data[k], dtype="object"
                )
            pd.DataFrame(df).to_csv(path_csv, index=False)
        else:
            pd.DataFrame(data).to_csv(path_csv, index=False, mode="a", header=False)


def combine_data_points_for_batch(data):
    # what could be the generic logic???
    if isinstance(data, list):
        joined = None
        for idx in range(len(data)):
            elem = data[idx]
            if isinstance(elem, dict):
                if joined is None:
                    joined = {}
                for key in list(elem.keys()):
                    if key not in joined:
                        joined.update({key: []})
                    joined[key].append(elem[key])
                    if idx == len(data) - 1:
                        joined[key] = np.squeeze(np.array(joined[key]), axis=1)
            elif isinstance(elem, list):
                if len(elem) == 1:
                    if joined is None:
                        joined = []
                    joined.append(np.array(elem))
                    if idx == len(data) - 1:
                        joined = np.squeeze(np.array(joined), axis=1)
                else:
                    print(data)
                    import pdb

                    pdb.set_trace()
                    raise Exception("Not implemented, please contact developers")
            elif isinstance(elem, np.ndarray):
                if joined is None:
                    joined = []
                joined.append(elem)
                joined = np.squeeze(np.array(joined), axis=1)
        return joined
    else:
        print(data)
        raise Exception("Not implemented, please contact developers")


def extract_data_points_from_batch(data, idxs):
    if isinstance(data, dict):
        this_data = {}
        for key in list(data.keys()):
            this_data.update({key: extract_data_points_from_batch(data[key], idxs)})
        return this_data
    elif isinstance(data, np.ndarray):
        return np.array(data[np.array(idxs)])
    elif isinstance(data, list):
        if isinstance(idxs, int):
            return data[idxs]
        else:
            return [data[x] for x in list(idxs)]
    else:
        return data


def make_2d_np_array(data):
    data = np.asarray(data)
    if len(data.shape) == 1:
        return np.expand_dims(data, axis=0)
    elif len(data.shape) == 2:
        return data
    else:
        raise Exception("Invalid data shape: %s" % str(data.shape))


def get_feature_names_list(inputs):
    feat_name_list = []
    for k in list(inputs.keys()):
        if not (k == "id"):
            feat_name_list.append(k)
    return feat_name_list


def add_data_to_batch(data, this_data):
    """Add data point this_data to data"""
    if isinstance(data, dict):
        assert isinstance(this_data, dict)
        for key, val in this_data.items():
            data.update({key: val})
        return data
    else:
        Exception("Invalid Data id type: %s" % type(data))


def get_df_indices_from_ids(df, ids):
    if not isinstance(df["id"][0], str):
        all_id_array = np.array(df["id"])
        if np.all(np.diff(all_id_array) >= 0):
            return np.searchsorted(all_id_array, ids)
        else:
            sorter = np.argsort(all_id_array)
            return sorter[np.searchsorted(all_id_array, ids, sorter=sorter)]
    else:
        id_str_list = [eval(x) for x in df["id"]]
        sorter = np.argsort(id_str_list)
        return sorter[np.searchsorted(id_str_list, ids, sorter=sorter)]


def read_json(file_name):
    with open(file_name) as f:
        data = json.load(f)
    return data


def write_json(file_name, data):
    with open(file_name, "w") as f:
        json.dump(data, f)


def write_csv_row(file_name, data):
    with open(file_name, "a") as f_object:
        writer_object = csv.writer(f_object)
        writer_object.writerow(data)
        f_object.close()


def read_csv(file_name):
    return pd.read_csv(file_name)


def load_list_from_df(df, column):
    try:
        out = [json.loads(x) for x in list(df[column])]
    except:
        out_json = [json.dumps(x) for x in list(df[column])]
        out = [json.loads(x) for x in out_json]
    return out


def make_dir_friendly_name(txt: str) -> str:
    import re

    return re.sub(r"[^a-zA-Z0-9_]", "_", txt)


def clear_directory(dir_path: str):
    """Clears the directory at dir_path but without deleting the directory itself. `shutil.rmtree` will
    have difficulties with mounted volumes or network drives.
    """
    import shutil

    for filename in os.listdir(dir_path):
        file_path = os.path.join(dir_path, filename)
        if os.path.isfile(file_path) or os.path.islink(file_path):
            os.unlink(file_path)
        elif os.path.isdir(file_path):
            shutil.rmtree(file_path)


class Clock:
    """Makes testing easier by anchoring to an older time. This is helpful when
    testing with older datasets. While if initialized with no arguments, the clock
    starts at the current time.
    """

    behind_by: timedelta
    tzone: Optional[tzinfo]

    def __init__(self, init_at: datetime):
        tz = init_at.tzinfo
        if tz is None or tz.utcoffset(init_at) is None:
            self.tzone = None
            self.behind_by = datetime.now() - init_at
        else:
            self.tzone = tz
            self.behind_by = datetime.now(tz=tz) - init_at

    def now(self) -> datetime:
        """Return the current time, adjusted by the amount of time the clock is behind."""
        return datetime.now(tz=self.tzone) - self.behind_by

    def sleep(self, seconds: float):
        """If the clock is behind, catch up. Else, sleep for the given duration."""
        seconds_behind = self.behind_by.total_seconds()
        if seconds_behind > 0:
            print(f"advancing the clock by {seconds} seconds")
            self.behind_by = timedelta(seconds=max(seconds_behind - seconds, 0))
        else:
            print("sleeping for 60 seconds")
            time.sleep(seconds)


class Timer:
    """Context manager for timing code blocks"""

    time: float

    def __enter__(self):
        self.time = time.perf_counter()
        return self

    def __exit__(self, type, value, traceback):
        self.time = round(time.perf_counter() - self.time, 3)


def dependency_required(dependency_name, package_name: str):
    """Decorator for checks that need optional dependencies. If the dependency is not
    present, initializing the class will raise an error.
    """

    def class_decorator(cls):
        @functools.wraps(cls, updated=())
        class WrappedClass(cls):
            def __init__(self, *args, **kwargs):
                if dependency_name is None:
                    raise ImportError(
                        f"{package_name} is required to use {cls.__name__}. Please install it using: pip install {package_name}"
                    )
                super().__init__(*args, **kwargs)

        return WrappedClass

    return class_decorator


def fn_dependency_required(dependency_name, package_name: str):
    """Decorator for functions that need optional dependencies. If the dependency is not
    present, calling the function will raise an error.
    """

    def fn_decorator(fn):
        @functools.wraps(fn)
        def wrapped_fn(*args, **kwargs):
            if dependency_name is None:
                raise ImportError(
                    f"{package_name} is required to use {fn.__name__}. Please install it using: pip install {package_name}"
                )
            return fn(*args, **kwargs)

        return wrapped_fn

    return fn_decorator



"""
Not used currently, refactored version of distribution.py
"""

from __future__ import annotations
from typing import TYPE_CHECKING, Union
import numpy as np
import random

from uptrain.v0.core.lib.helper_funcs import make_2d_np_array
from uptrain.v0.core.classes.distances import DistanceResolver
from uptrain.v0.core.classes.statistics import AbstractStatistic
from uptrain.v0.core.classes.measurables import MeasurableResolver
from uptrain.v0.constants import Statistic
from uptrain.v0.core.lib.cache import make_cache_container

if TYPE_CHECKING:
    from uptrain.v0.core.classes.logging.new_log_handler import (
        LogHandler as NewLogHandler,
        LogWriter,
    )
    from uptrain.v0.core.classes.logging.log_handler import LogHandler


class Distribution(AbstractStatistic):
    log_handler: Union["LogHandler", "NewLogHandler"]
    log_writers: list["LogWriter"]  # one for each distance type
    dashboard_name = "distribution_stats"
    statistic_type = Statistic.DISTRIBUTION_STATS
    bucket_size = 400  # max num of embeddings sampled at each checkpoint
    refresh_rate = 0.1  # churn from the previous set at each checkpoint

    def base_init(self, fw, check):
        self.aggregate_measurable = MeasurableResolver(check["aggregate_args"]).resolve(
            fw
        )
        self.count_measurable = MeasurableResolver(check["count_args"]).resolve(fw)
        self.distance_types = check["distance_types"]
        self.count_checkpoints = list(sorted(set([0] + check["count_checkpoints"])))

        # setup a dict to store the sampled embeddings at each checkpoint
        self.values_for_cp = {cp: [] for cp in self.count_checkpoints}

        # setup a cache to store interim state for aggregates
        attrs_to_store = {"prev_count": np.ndarray, "first_checkpoint": int}
        self.cache = make_cache_container(fw, attrs_to_store)

        self.dist_classes = [DistanceResolver().resolve(x) for x in self.distance_types]
        if hasattr(self.log_handler, "make_logger"):
            self.log_writers = [
                self.log_handler.make_logger(self.dashboard_name, distance_type)
                for distance_type in self.distance_types
            ]  # get handles to log writers for each distance type
        else:
            self.log_writers = []

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        all_models = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.model_measurables
        ]
        all_features = [
            x.compute_and_log(inputs, outputs, gts=gts, extra=extra_args)
            for x in self.feature_measurables
        ]
        vals = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        aggregate_ids = self.aggregate_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )
        counts = self.count_measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )

        # new embeddings seen for each checkpoint
        new_values_for_cp = {cp: [] for cp in self.count_checkpoints}

        # read previous state from the cache
        # - prev_count_cache: the previous view count seen for each aggregate id
        # - first_crossed_checkpoint_cache: the first checkpoint that was crossed for each aggregate id
        [
            prev_count_cache,
            first_crossed_checkpoint_cache,
        ] = self.cache.fetch_col_values_for_ids(
            np.asarray(aggregate_ids),
            ["prev_count", "first_checkpoint"],
        )

        set_ids_to_cache = set()  # track which ids we need to update in the cache
        for idx in range(len(aggregate_ids)):
            is_model_invalid = sum(
                [
                    all_models[jdx][idx] not in self.allowed_model_values[jdx]
                    for jdx in range(len(self.allowed_model_values))
                ]
            )
            if is_model_invalid:
                continue

            active_id = aggregate_ids[idx]
            prev_count = prev_count_cache.get(aggregate_ids[idx], None)
            curr_count = counts[idx]

            # Four possible cases:
            # 1. prev_count = None: first time we see this aggregate id, so we skip processing it. Save the current state in the cache.
            # 2. prev_count > max(checkpoints): we skip processing this row altogether since all checkpoints have been crossed.
            # 3. prev_count > curr_count: this row is out of order, so we skip processing it.
            # 4. curr_count > prev_count: we have seen this aggregate id before, so we process it. Save the current state in the cache.
            if prev_count is None:
                crossed_checkpoint = max(
                    cp for cp in self.count_checkpoints if cp <= curr_count
                )
                set_ids_to_cache.add(active_id)
                prev_count_cache[active_id] = curr_count
                first_crossed_checkpoint_cache[active_id] = crossed_checkpoint
            elif prev_count > self.count_checkpoints[-1]:
                continue
            elif prev_count > curr_count:
                continue
            else:
                last_crossed_checkpoint = None
                for cp in reversed(self.count_checkpoints):
                    if prev_count > cp:
                        break  # we are already past this checkpoint or those smaller than it
                    elif (prev_count <= cp) and (cp < curr_count):
                        last_crossed_checkpoint = cp
                        break

                set_ids_to_cache.add(active_id)
                prev_count_cache[active_id] = curr_count
                if last_crossed_checkpoint is None:
                    # no new checkpoint crossed, so we just update the last count seen and move on
                    continue
                else:
                    # new checkpoint crosses, so add this embedding to the list for this checkpoint
                    new_values_for_cp[last_crossed_checkpoint].append(vals[idx])

        # update samples for each checkpoint and compute distributions
        for cp in self.count_checkpoints:
            num_new_samples = len(new_values_for_cp[cp])
            num_to_add = int(self.refresh_rate * self.bucket_size)
            if num_to_add > num_new_samples:
                # we have enough new samples to replace the old ones
                self.values_for_cp[cp] += new_values_for_cp[cp]
            else:
                # we have too many new samples, so randomly choose from them
                self.values_for_cp[cp] += random.sample(
                    new_values_for_cp[cp], num_to_add
                )
            self.values_for_cp[cp] = self.values_for_cp[cp][-self.bucket_size :]

            # samples N pairs from these embedding values and compute distance between them
            num_pairs_to_sample = 400
            total_num_samples = len(self.values_for_cp[cp])

            first_indices = np.random.choice(total_num_samples, num_pairs_to_sample)
            second_indices = np.random.choice(total_num_samples, num_pairs_to_sample)
            invalid_pairs = first_indices == second_indices
            second_indices[invalid_pairs] = (second_indices[invalid_pairs] + 1) % total_num_samples  # type: ignore

            first_vector = np.vstack(
                [self.values_for_cp[cp][idx] for idx in first_indices]
            )
            second_vector = np.vstack(
                [self.values_for_cp[cp][idx] for idx in second_indices]
            )
            for k, dist_class in enumerate(self.dist_classes):
                distances = dist_class.compute_distance(first_vector, second_vector)
                # TODO: are we creating separate histograms for each model, feature type pair?

        # update the cache with data from this batch
        if len(set_ids_to_cache) > 0:
            # save the current state in the cache
            list_ids = list(set_ids_to_cache)
            list_counts, list_first_checkpoints = [], []
            for _id in list_ids:
                list_counts.append(prev_count_cache[_id])
                list_first_checkpoints.append(first_crossed_checkpoint_cache[_id])

            self.cache.upsert_ids_n_col_values(
                np.asarray(list_ids),
                {
                    "prev_count": np.asarray(list_counts),
                    "first_checkpoint": np.asarray(list_first_checkpoints),
                },
            )



"""
This module provides the Topic Assignment operators that help to assign topics based on input queries.

Operators:
- `TopicAssignmentviaCluster`: Assigns topics based on clustering performed on embeddings.

"""

from __future__ import annotations
import typing as t

from loguru import logger
import numpy as np
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep


@register_op
class TopicAssignmentviaCluster(ColumnOp):

    """
    Operator for assigning topics based on cluster assignments. Note, you should run Clustering operator before using this.

    Attributes:
        cluster_centroids (dict): Dictionary of centroids for each cluster for each aggregation. Extracted from 'Clustering' operator.
        topics (dict): Dictionary of topics for each cluster for each aggregation. Extracted from 'TopicGeneration' operator.
        col_embeddings (str):  The name of the column which contains embeddings - It will be compared against centroids to determine cluster.
        col_out (str):  The name of the column in the DataFrame to output the assigned topic.
        col_out_cluster (str): The name of the column in the DataFrame to output the assigned cluster index.
        col_out_dist (str): The name of the column in the DataFrame to output the euclidean distance from its cluster centroid.
        col_aggs (list[str]): Optional,pecify name of columns to aggregate by which was used during clustering, ex: if you ran separate clustering for seperate organisations
    """

    cluster_centroids: dict
    topics: dict
    col_embeddings: str = 'embedding'
    col_out: str = 'topic'
    col_out_cluster: str = "cluster_index"
    col_out_dist: str = 'cluster_index_distance'
    col_aggs: list[str] = []

    def setup(self, settings: Settings):
        assert len(self.topics) > 0, "Topic list should not be empty"
        assert len(self.topics) == len(self.cluster_centroids), "Each group should have a topic"
        for key in self.topics.keys():
            assert len(self.topics[key]) == len(self.cluster_centroids[key]), "Each cluster should have a topic"
            self.cluster_centroids[key] = np.array(self.cluster_centroids[key])
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:      
        res_data_arr = []

        unique_agg_keys = ['default']
        if len(self.col_aggs):
            # First aggregate by col_aggs
            agg_data = data.groupby(self.col_aggs).agg([pl.col(self.col_embeddings).count().alias("num_rows_" + self.col_embeddings)])
            agg_data = agg_data.drop("num_rows_" + self.col_embeddings)
            unique_agg_keys = agg_data.to_dicts()

        for unique_agg_key in unique_agg_keys:
            cond = True
            if isinstance(unique_agg_key, dict):
                unique_agg_key = dict([(key, unique_agg_key[key]) for key in sorted(unique_agg_key)])
                for key,val in unique_agg_key.items():
                    cond = cond & (data[key] == val)
            data_subset = data.filter(cond)
            embeddings = np.asarray(data_subset[self.col_embeddings])

            if str(unique_agg_key) not in self.cluster_centroids:
                cluster_centroids = None
                topics = ['Not Defined']
            else:
                cluster_centroids = np.array(self.cluster_centroids[str(unique_agg_key)])
                topics = self.topics[str(unique_agg_key)]

            assigned_topics = []
            assigned_clusters = []
            cluster_index_distances = []
            for index in range(len(embeddings)):
                if cluster_centroids is None:
                    cluster_index_distances.append(float(-1))
                    assigned_topics.append(topics[0])
                    assigned_clusters.append(-1)
                else:
                    dists = np.linalg.norm(embeddings[index] - cluster_centroids, axis=1)
                    assigned_cluster = np.argmin(dists)
                    cluster_index_distances.append(np.min(dists))
                    assigned_topics.append(topics[assigned_cluster])
                    assigned_clusters.append(assigned_cluster)

            data_subset = data_subset.with_columns([
                    pl.Series(assigned_topics).alias(self.col_out),
                    pl.Series(assigned_clusters).alias(self.col_out_cluster),
                    pl.Series(cluster_index_distances).alias(self.col_out_dist),
            ])
            res_data_arr.append(data_subset)

        return {
            "output": pl.concat(res_data_arr)
        }




from typing import Any, Dict

from uptrain.v0.core.classes.measurables import Measurable, FeatureMeasurable


class ConditionMeasurable(Measurable):
    """Class that computes a boolean condition on a measurable feature."""

    def __init__(
        self, framework, underlying: Dict[str, Any], condition: Dict[str, Any]
    ) -> None:
        """Constructor for ConditionMeasurable class.

        Parameters
        ----------
        framework
            UpTrain Framework object
        underlying
            A dictionary containing the feature_name to apply the condition on and
            the dictn_type (whether to apply on inputs or outputs).
            dictn_type should be "inputs" for applying InputFeatureMeasurable
            dictn_type should be "outputs" for applying OutputFeatureMeasurable
        condition
            A dictionary containing either "formulae" or "func" as key.
            If "formulae" is provided:
                - it must be a boolean operator from the following list of operators:
                    ["<", "<=", ">", ">=", "==", "le", "leq", "ge", "geq", "eq"]
                - it must contain "threshold" as key which provides the value to perform
                  the boolean condition against.
            If "func" is provided, it should map to a callback that accepts a single
            argument and returns a boolean value.
        """
        super().__init__(framework)
        self.underlying = underlying
        self.condition = condition
        self.feature = FeatureMeasurable(
            framework, underlying["feature_name"], underlying["dictn_type"]
        )
        if "formulae" in condition:
            formulae = condition["formulae"]
            if (formulae == "leq") or (formulae == "<="):
                condition_func = lambda x: (x <= condition["threshold"])
            elif (formulae == "le") or (formulae == "<"):
                condition_func = lambda x: (x < condition["threshold"])
            elif (formulae == "geq") or (formulae == ">="):
                condition_func = lambda x: (x >= condition["threshold"])
            elif (formulae == "ge") or (formulae == ">"):
                condition_func = lambda x: (x > condition["threshold"])
            elif (formulae == "eq") or (formulae == "=="):
                condition_func = lambda x: (x == condition["threshold"])
            else:
                raise Exception("Condition Formulae %s is not supported" % formulae)
        elif "func" in condition:
            condition_func = condition["func"]
        else:
            raise Exception(
                "Either func or formulae should be defined to resolve condition"
            )
        self.condition_func = condition_func

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> Any:
        return self.condition_func(
            self.feature._compute(inputs=inputs, outputs=outputs, gts=gts, extra=extra)
        )

    def col_name(self) -> str:
        return f"Condition({self.feature.col_name()}, {str(self.condition)})"



"""
This module provides the following Distribution and UMAP operators that operate on embeddings:

Operators:
- `Distribution`: Computes the distribution of similarity metrics.
- `UMAP`: Performs UMAP dimensionality reduction.

NOTE: The above operators only take embeddings as input. Refer uptrain.operators.language.embedding to learn more.

"""

from __future__ import annotations
import typing as t

from loguru import logger
import numpy as np
import polars as pl
from pydantic import root_validator

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep

umap = lazy_load_dep("umap", "umap-learn")
rouge_scorer = lazy_load_dep("rouge_score.rouge_scorer", "rouge_score")


@register_op
class Distribution(TransformOp):
    """
    Operator for computing distribution of similarity metrics.

    Attributes:
        kind (Literal["cosine_similarity", "rouge"]): The type of similarity metric.
        col_in_embs (list[str]): The input columns containing embeddings.
        col_in_groupby (list[str]): The columns to group by.
        col_out (list[str] | None): The output columns. If None, automatically generated column names will be used.

    Raises:
        AssertionError: If the number of output columns does not match the number of input embedding columns.

    Example:
        ```
        import polars as pl
        from uptrain.operators import Distribution

        # Create an instance of the Distribution operator
        op = Distribution(
                kind="cosine_similarity",
                col_in_embs=["context_embeddings", "response_embeddings"],
                col_in_groupby=["question_idx", "experiment_id"],
                col_out=["similarity-context", "similarity-response"],
            )

        # Set up the operator
        op.setup()

        # Run the operator on the input data
        input_data = pl.DataFrame(...)
        output = op.run(input_data)["output"]

        # Print the output
        print(output)
        ```


    Output:
        ```
        shape: (90, 4)
        ┌──────────────┬───────────────┬────────────────────┬─────────────────────┐
        │ question_idx ┆ experiment_id ┆ similarity-context ┆ similarity-response │
        │ ---          ┆ ---           ┆ ---                ┆ ---                 │
        │ i64          ┆ i64           ┆ f64                ┆ f64                 │
        ╞══════════════╪═══════════════╪════════════════════╪═════════════════════╡
        │ 2            ┆ 0             ┆ 0.314787           ┆ 1.0                 │
        │ 2            ┆ 0             ┆ 0.387398           ┆ 0.204949            │
        │ 2            ┆ 0             ┆ 0.344797           ┆ 0.23195             │
        │ 2            ┆ 0             ┆ 0.306041           ┆ 1.0                 │
        │ …            ┆ …             ┆ …                  ┆ …                   │
        │ 0            ┆ 2             ┆ 0.997804           ┆ 0.996358            │
        │ 0            ┆ 2             ┆ 0.66862            ┆ 0.300155            │
        │ 0            ┆ 2             ┆ 0.224229           ┆ 0.637781            │
        │ 0            ┆ 2             ┆ 0.379936           ┆ 0.260659            │
        └──────────────┴───────────────┴────────────────────┴─────────────────────┘
        ```

    """

    kind: t.Literal["cosine_similarity", "rouge"]
    col_in_embs: list[str]
    col_in_groupby: list[str]
    col_out: list[str] | None = None

    @root_validator(pre=True)
    def _check_cols(cls, values):
        """
        Validator to check the validity of input and output column lists.

        Args:
            values (dict): The input attribute values.

        Returns:
            dict: The validated attribute values.

        Raises:
            AssertionError: If the number of output columns does not match the number of input embedding columns.

        """
        if values["col_out"] is not None:
            assert len(values["col_out"]) == len(
                values["col_in_embs"]
            ), "Distribution Op needs as many output columns as input embedding columns"
        return values

    def setup(self, settings: Settings):
        if self.kind == "cosine_similarity":
            self._agg_func = get_cosine_sim_dist
        elif self.kind == "rouge":
            self._agg_func = get_rouge_score
        else:
            raise NotImplementedError(
                f"Similarity metric: {self.kind} not supported for now."
            )
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        if self.col_out is None:
            agg_cols = [get_output_col_name_at(i) for i in range(len(self.col_in_embs))]
        else:
            agg_cols = self.col_out

        dist_df = (
            data.groupby(self.col_in_groupby, maintain_order=True)
            .agg(
                [
                    pl.col(_col_in).apply(self._agg_func).alias(_col_out)
                    for _col_in, _col_out in zip(self.col_in_embs, agg_cols)
                ]
            )
            .explode(agg_cols)
        )
        return {"output": dist_df}


@register_op
class UMAP(ColumnOp):
    """
    Operator for performing UMAP dimensionality reduction.

    Attributes:
        col_in_embs (str): The input column containing embeddings.
        n_components (int): Number of components to keep.
        col_out (str): The umap column containing embeddings. 

    Example:
        ```
        import polars as pl
        from uptrain.operators import UMAP

        # Create an instance of the UMAP operator
        op = UMAP(
                col_in_embs="embedding",
                n_components=6,
                col_out= "umap_embedding"

            )

        # Set up the operator
        op.setup()

        # Run the operator on the input data
        input_data = pl.DataFrame(...)
        output = op.run(input_data)

        # Get the output DataFrame
        umap_df = output["output"]
        ```

    Output:
        ```
        shape: (2,)
        Series: '_col_0' [list[f32]]
        [
                [0.098575, 0.056978, … -0.071038]
                [0.072772, 0.073564, … -0.043947]
        ]
        ```

    """

    col_in_embs: str = 'embedding'
    n_components: int = 6
    col_out: str = 'umap_embedding' 

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        combined_embs = np.asarray(list(data[self.col_in_embs]))
        umap_output = umap.UMAP(n_components=self.n_components, metric='cosine', random_state=42).fit_transform(combined_embs)  # type: ignore
        umap_output = [[round(y, 8) for y in list(x)] for x in list(umap_output)]
        output_cols = [pl.Series(umap_output).alias(self.col_out)]
        output_cols.extend([pl.Series(np.array(umap_output)[:,idx]).alias(self.col_out + "_" + str(idx)) for idx in range(self.n_components)])

        return {
            "output": data.with_columns(output_cols)
        }


# -----------------------------------------------------------
# Utility routines (for above operators)
# -----------------------------------------------------------


def sample_pairs_from_values(n_values: int, n_pairs: int):
    """
    Sample pairs of indices from a given number of values.

    Args:
        n_values (int): The total number of values.
        n_pairs (int): The number of pairs to sample.

    Returns:
        Tuple[np.ndarray, np.ndarray]: The sampled pairs of indices.

    """
    indices_1 = np.random.choice(n_values, n_pairs)
    indices_2 = np.random.choice(n_values, n_pairs)
    invalid = indices_1 == indices_2
    indices_2[invalid] = (indices_2[invalid] + 1) % n_values
    return indices_1, indices_2


def get_cosine_sim_dist(col_vectors: pl.Series, num_pairs_per_group: int = 10):
    """
    Compute cosine similarity distances between pairs of vectors.

    Args:
        col_vectors (pl.Series): The column containing the vectors.
        num_pairs_per_group (int): The number of pairs to sample per group.

    Returns:
        List[float]: The computed cosine similarity distances.

    """
    array_vectors = col_vectors.to_numpy()
    indices_1, indices_2 = sample_pairs_from_values(
        len(array_vectors), num_pairs_per_group
    )
    values = []
    for i1, i2 in zip(indices_1, indices_2):
        v1 = array_vectors[i1]
        v2 = array_vectors[i2]
        values.append(np.dot(v1, v2) / np.linalg.norm(v1) * np.linalg.norm(v2))
    return values


def get_rouge_score(col_vectors: pl.Series, num_pairs_per_group: int = 10):
    """
    Compute ROUGE scores between pairs of vectors.

    Args:
        col_vectors (pl.Series): The column containing the vectors.
        num_pairs_per_group (int): The number of pairs to sample per group.

    Returns:
        List[int]: The computed ROUGE scores.

    """
    array_vectors = col_vectors.to_numpy()
    indices_1, indices_2 = sample_pairs_from_values(
        len(array_vectors), num_pairs_per_group
    )
    values = []
    for i1, i2 in zip(indices_1, indices_2):
        v1 = array_vectors[i1]
        v2 = array_vectors[i2]

        scorer = rouge_scorer.RougeScorer(["rougeL"])  # type: ignore
        values.append(int(scorer.score(v1, v2)["rougeL"][2] * 100))
    return values



"""Base constructs used to create user-facing objects.
"""

from __future__ import annotations
import os
import typing as t

from loguru import logger
import networkx as nx
import polars as pl
from pydantic import BaseSettings, Field

from uptrain.operators.base import *
from uptrain.utilities import to_py_types, jsondump, jsonload

__all__ = [
    "OperatorDAG",
    "Settings",
]


class Settings(BaseSettings):
    # uptrain stores logs in this folder
    logs_folder: str = "/tmp/uptrain-logs"

    # external api related
    openai_api_key: str = Field(None, env="OPENAI_API_KEY")
    cohere_api_key: str = Field(None, env="COHERE_API_KEY")
    huggingface_api_key: str = Field(None, env="HUGGINGFACE_API_KEY")
    anthropic_api_key: str = Field(None, env="ANTHROPIC_API_KEY")
    replicate_api_token: str = Field(None, env="REPLICATE_API_TOKEN")

    azure_api_key: str = Field(None, env="AZURE_API_KEY")
    azure_api_base: str = Field(None, env="AZURE_API_BASE")
    azure_api_version: str = Field(None, env="AZURE_API_VERSION")

    rpm_limit: int = 100
    tpm_limit: int = 90_000
    embedding_compute_method: t.Literal['local', 'replicate', 'api'] = 'local'

    # uptrain managed service related
    uptrain_access_token: str = Field(None, env="UPTRAIN_ACCESS_TOKEN")
    uptrain_server_url: str = Field("https://demo.uptrain.ai/", env="UPTRAIN_SERVER_URL")

    # Embedding model related, applicable if embedding_compute_method is api.
    embedding_model_url: str = Field(None, env="EMBEDDING_MODEL_URL")
    embedding_model_api_token: str = Field(None, env="EMBEDDING_MODEL_API_TOKEN")

    # LLM model to run the evaluations
    model: str = "gpt-3.5-turbo-1106"
    seed:  t.Union[int, None] = None
    response_format: t.Union[dict, None] = None

    # allow additional fields as needed by different operators
    class Config:
        extra = "allow"


    def __init__(self, **data):
        super().__init__(**data)
        if "openai_api_key" in data:
            if data['openai_api_key'] is not None:
                os.environ["OPENAI_API_KEY"] = data["openai_api_key"]
        if "cohere_api_key" in data:
            if data['cohere_api_key'] is not None:
                os.environ["COHERE_API_KEY"] = data["cohere_api_key"]
        if "huggingface_api_key" in data:
            if data['huggingface_api_key'] is not None:
                os.environ["HUGGINGFACE_API_KEY"] = data["huggingface_api_key"]
        if "anthropic_api_key" in data:
            if data['anthropic_api_key'] is not None:
                os.environ["ANTHROPIC_API_KEY"] = data["anthropic_api_key"]
        if "replicate_api_token" in data:
            if data['replicate_api_token'] is not None:
                os.environ["REPLICATE_API_TOKEN"] = data["replicate_api_token"]
        if "embedding_model_api_token" in data:
            if data['embedding_model_api_token'] is not None:
                os.environ["EMBEDDING_MODEL_API_TOKEN"] = data["embedding_model_api_token"]
        if "azure_api_key" in data:
            if data['azure_api_key'] is not None:
                os.environ["AZURE_API_KEY"] = data["azure_api_key"]
        if "azure_api_base" in data:
            if data['azure_api_base'] is not None:
                os.environ["AZURE_API_BASE"] = data["azure_api_base"]
        if "azure_api_version" in data:
            if data['azure_api_version'] is not None:
                os.environ["AZURE_API_VERSION"] = data["azure_api_version"]
        if "uptrain_access_token" in data:
            if data['uptrain_access_token'] is not None:
                os.environ["UPTRAIN_ACCESS_TOKEN"] = data["uptrain_access_token"]
        if "uptrain_server_url" in data:
            if data['uptrain_server_url'] is not None:
                os.environ["UPTRAIN_SERVER_URL"] = data["uptrain_server_url"]

    def check_and_get(self, key: str) -> t.Any:
        """Check if a value is present in the settings and return it."""
        value = getattr(self, key)
        if value is None:
            raise ValueError(f"Expected value for {key} to be present in the settings.")
        return value

    def serialize(self, fpath: str | None = None):
        """Serialize the settings to a json file."""
        if fpath is None:
            fpath = os.path.join(self.logs_folder, "settings.json")
        with open(fpath, "w") as f:
            jsondump(self.dict(), f)

    @classmethod
    def deserialize(cls, fpath: str):
        """Deserialize the settings from a json file."""
        with open(fpath, "r") as f:
            return cls(**jsonload(f))


class OperatorDAG:
    """A Graph is a DAG of Uptrain table operators, that defines the data pipeline to execute."""

    name: str
    graph: nx.DiGraph

    def __init__(self, name: str):
        self.name = name
        self.graph = nx.DiGraph()

    def add_step(
        self, name: str, node: Operator, deps: t.Optional[list[str]] = None
    ) -> None:
        """Add a node to the DAG, along with its dependencies."""
        if name in self.graph.nodes:
            raise ValueError(
                f"Operator with name: {name} already exists in the compute DAG: {self.name}"
            )
        self.graph.add_node(name, op_class=node)

        if deps is None:
            deps = []
        for dep in deps:
            if dep not in self.graph.nodes:
                raise ValueError(
                    f"Specified dependency: {dep} does not exist in the compute DAG: {self.name}"
                )
            self.graph.add_edge(dep, name)

        if not nx.algorithms.dag.is_directed_acyclic_graph(self.graph):
            self.graph.remove_node(name)
            raise ValueError(
                f"Adding a node for operator: {name} creates a cycle in the compute DAG: {self.name}"
            )

    def setup(self, settings: "Settings") -> None:
        """Set up the operators in the DAG."""
        sorted_nodes = list(nx.algorithms.dag.topological_sort(self.graph))
        for node_name in sorted_nodes:
            node: "Operator" = self.graph.nodes[node_name]["op_class"]
            node.setup(settings)

    def run(
        self,
        node_inputs: dict[str, pl.DataFrame | None],
        output_nodes: list[str],
    ) -> dict[str, pl.DataFrame]:
        """Runs the compute DAG.
        Args:
            node_inputs: A dict of input dataframes, keyed by operator name. For other operators, the output
                from the upstream operators is used as input.
            node_outputs: A list of operator names, whose output should be returned.
        """

        # dict to hold the output of each node
        node_to_output = {}
        sorted_nodes = list(nx.algorithms.dag.topological_sort(self.graph))
        dependents_count = {
            node_name: len(self._get_node_children(node_name))
            for node_name in sorted_nodes
        }

        # run each node in topological order
        for node_name in sorted_nodes:
            logger.debug(f"Executing node: {node_name} for operator DAG: {self.name}")
            node: "TransformOp" = self.graph.nodes[node_name]["op_class"]

            # get input for this node from its dependencies
            inputs_from_deps = []
            if node_name in node_inputs:
                inputs_from_deps.append(node_inputs[node_name])
            else:
                for dep in self.graph.predecessors(node_name):
                    if dep in node_to_output:
                        inputs_from_deps.append(node_to_output[dep])
                    else:
                        raise ValueError(
                            f"Cannot find output/provided value for dependency: {dep} of node: {node_name}"
                        )

            # run the operator and store the output
            res = node.run(*inputs_from_deps)
            node_to_output[node_name] = res["output"]

            # decrease dependents count for each dependency so we don't old onto memory
            for parent in self.graph.predecessors(node_name):
                dependents_count[parent] -= 1
                if dependents_count[parent] == 0 and parent not in output_nodes:
                    node_to_output.pop(parent, None)

        return {node_name: node_to_output[node_name] for node_name in output_nodes}

    def _get_node_parents(self, name: str) -> list[str]:
        return list(self.graph.predecessors(name))

    def _get_node_children(self, name: str) -> list[str]:
        return list(self.graph.successors(name))

    def __repr__(self) -> str:
        sorted_nodes = list(nx.algorithms.dag.topological_sort(self.graph))
        lines = []
        for node_name in sorted_nodes:
            lines.append(f"{node_name} <- {self._get_node_parents(node_name)}")
        return f"Check: {self.name}\n" + "\n".join(lines)

    def dict(self) -> dict:
        """Serialize this check to a dict."""
        nodes = {
            name: to_py_types(self.graph.nodes[name]["op_class"])
            for name in self.graph.nodes
        }
        edges = list(self.graph.edges)
        return {"name": self.name, "nodes": nodes, "edges": edges}

    @classmethod
    def from_dict(cls, data: dict) -> "OperatorDAG":
        """Deserialize a check from a dict."""
        op_dag = cls(name=data["name"])
        for name, node in data["nodes"].items():
            op_dag.graph.add_node(name, op_class=deserialize_operator(node))
        for edge in data["edges"]:
            assert edge[0] in op_dag.graph.nodes, f"Operator node: {edge[0]} not found"
            assert edge[1] in op_dag.graph.nodes, f"Operator node: {edge[1]} not found"
            op_dag.graph.add_edge(edge[0], edge[1])
        return op_dag



import numpy as np
from uptrain.v0.core.classes.monitors import AbstractMonitor
from uptrain.v0.core.classes.measurables import MeasurableResolver
from uptrain.v0.constants import Monitor, ComparisonModel, ComparisonMetric


class OutputComparison(AbstractMonitor):
    dashboard_name = "output_comparison"
    monitor_type = Monitor.OUTPUT_COMPARISON

    def base_init(self, fw, check):
        self.comparison_model_base = check["comparison_model"]
        self.comparison_model_resolved = ComparisonModelResolver().resolve(
            check["comparison_model"]
        )
        self.comparison_model_inputs = MeasurableResolver(
            check.get("comparison_model_input_args", None)
        ).resolve(fw)
        self.comparison_metric_base = check["comparison_metric"]
        self.comparison_metric_resolved = ComparisonMetricResolver().resolve(
            check["comparison_metric"]
        )
        self.threshold = check["threshold"]
        self.count = 0

    def base_check(self, inputs, outputs, gts=None, extra_args={}):
        vals = self.measurable.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )

        comparison_model_inputs = self.comparison_model_inputs.compute_and_log(
            inputs, outputs, gts=gts, extra=extra_args
        )

        comparison_model_outputs = self.comparison_model_resolved(
            comparison_model_inputs
        )
        batch_metrics = self.comparison_metric_resolved(vals, comparison_model_outputs)
        self.batch_metrics = batch_metrics

        extra_args.update(
            {
                self.comparison_model_base + " outputs": comparison_model_outputs,
                self.comparison_metric_base: batch_metrics,
            }
        )

        feat_name = self.comparison_metric_base
        plot_name = (
            f"{feat_name} Comparison - Production vs {self.comparison_model_base}"
        )
        self.count += len(extra_args["id"])

        self.log_handler.add_scalars(
            plot_name,
            {"y_" + feat_name: np.mean(batch_metrics)},
            self.count,
            self.dashboard_name,
            file_name=plot_name,
        )

    def need_ground_truth(self):
        return False

    def base_is_data_interesting(self, inputs, outputs, gts=None, extra_args={}):
        reasons = ["None"] * len(extra_args["id"])
        is_interesting = self.batch_metrics < self.threshold
        reasons = []
        for idx in range(len(extra_args["id"])):
            if is_interesting[idx] == 0:
                reasons.append("None")
            else:
                reasons.append(
                    f"Different output compared to {self.comparison_model_base}"
                )
        return is_interesting, reasons


class ComparisonModelResolver:
    def resolve(self, model):
        if model == ComparisonModel.FASTER_WHISPER:
            from uptrain.v0.ee.lib.algorithms import faster_whisper_speech_to_text

            return faster_whisper_speech_to_text
        else:
            raise Exception(f"{model} can't be resolved")


class ComparisonMetricResolver:
    def resolve(self, metric):
        if metric == ComparisonMetric.ROGUE_L_F1:
            from uptrain.v0.ee.lib.algorithms import rogue_l_similarity

            return rogue_l_similarity
        else:
            raise Exception(f"{metric} can't be resolved")



from __future__ import annotations

import argparse
import subprocess
import sys


EXTRA_PACKAGES = {
    "v0": [
        "pandas>=1.0.0",
        "river",
        "scikit-learn>=1.0.0",
        "umap-learn",
        "json-fix>=0.5.0",
        "xgboost",
    ],
    "local-generation": [
        "tqdm>=4.0",
        "aiolimiter>=1.1",
    ],
    "dashboard": [
        "plotly>=5.0.0",
        "streamlit>=1.23",
        "pyarrow>=10.0.0"
    ],
    "selfhosted": [
        "tqdm>=4.0",
        "aiolimiter>=1.1",
        "plotly>=5.0.0",
        "streamlit>=1.23",
        "pyarrow>=10.0.0",
        "river<0.18",
        "scikit-learn>=1.0.0",
        "umap-learn",
        "rouge-score",
        "torchmetrics",
        # "sentence-transformers",
        # "InstructorEmbedding",
        "openai==1.6.1",
        "llama-index>=0.8.0",
        "evals @ git+https://github.com/openai/evals.git@main",
        "replicate",
        "litellm==1.15.7",
    ],
    # Add packages here
    "full": [
        "tqdm>=4.0",
        "aiolimiter>=1.1",
        "plotly>=5.0.0",
        "streamlit>=1.23",
        "pyarrow>=10.0.0",
        "river<0.18",
        "scikit-learn>=1.0.0",
        "umap-learn",
        "rouge-score",
        "torchmetrics",
        "sentence-transformers",
        "InstructorEmbedding",
        "openai==1.6.1",
        "llama-index>=0.8.0",
        "evals @ git+https://github.com/openai/evals.git@main",
        "replicate",
        "litellm==1.15.7",
    ],
}


def make_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--feature",
        choices=EXTRA_PACKAGES.keys(),
        required=True,
        help="Install extra packages for using specific features in Uptrain.",
    )
    return parser


def main(argv=None) -> int:
    parser = make_parser()
    args = parser.parse_args(argv)

    if args.feature not in EXTRA_PACKAGES:
        print(f"Unknown feature: {args.feature}", file=sys.stderr)
        return 1
    else:
        packages = EXTRA_PACKAGES[args.feature]
        if args.feature == "full":
            # we want to install torch-cpu if we are installing the full feature
            # Nvidia packages blow up the size of the docker image
            subprocess.check_call(
                [
                    sys.executable,
                    "-m",
                    "pip",
                    "install",
                    "torch",
                    "torchvision",
                    "--index-url",
                    "https://download.pytorch.org/whl/cpu",
                ],
            )
        subprocess.check_call(
            [sys.executable, "-m", "pip", "install", *packages],
        )
        return 0



"""This module hosts built-in checks for some common LLM evaluation tasks."""

from __future__ import annotations
import typing as t

from .checks import Check
from uptrain.operators import Histogram
from uptrain.operators import (
    ResponseFactualScore,
    ValidResponseScore,
    ResponseCompleteness,
    ResponseRelevance,
    ResponseConsistency,
    ResponseConciseness,
    ContextRelevance,
    PromptInjectionScore,
    LanguageCritique,
    ToneCritique,
    ResponseCompletenessWrtContext,
    GuidelineAdherenceScore,
    ResponseMatchingScore, 
    ConversationSatisfactionScore
)

# -----------------------------------------------------------
# Context related
# -----------------------------------------------------------

CheckContextRelevance = lambda: Check(
    name="score_context_relevance",
    operators=[ContextRelevance()],
    plots=[Histogram(x="score_context_relevance")],
)

CheckResponseFacts = lambda: Check(
    name="score_factual_accuracy",
    operators=[ResponseFactualScore()],
    plots=[Histogram(x="score_factual_accuracy")],
)

# -----------------------------------------------------------
# Response related
# -----------------------------------------------------------

CheckResponseCompleteness = lambda: Check(
    name="response_completeness_score",
    operators=[ResponseCompleteness()],
    plots=[Histogram(x="score_response_completeness")],
)


CheckResponseCompletenessWrtContext = lambda: Check(
    name="response_completeness_wrt_context_score",
    operators=[ResponseCompletenessWrtContext()],
    plots=[Histogram(x="score_response_completeness_wrt_context")],
)


CheckResponseRelevance = lambda: Check(
    name="response_relevance_score",
    operators=[ResponseRelevance()],
    plots=[Histogram(x="score_response_relevance")],
)

CheckResponseConsistency = lambda: Check(
    name="response_consistency_score",
    operators=[ResponseConsistency()],
    plots=[Histogram(x="score_response_consistency")],
)

CheckValidResponse = lambda: Check(
    name="valid_response_score",
    operators=[ValidResponseScore()],
    plots=[Histogram(x="score_valid_response")],
)


CheckResponseConciseness = lambda: Check(
    name="response_conciseness_score",
    operators=[ResponseConciseness()],
    plots=[Histogram(x="score_response_conciseness")],
)
# -----------------------------------------------------------
# Language quality related
# -----------------------------------------------------------

CheckLanguageQuality = lambda: Check(
    name="language_critique_score",
    operators=[LanguageCritique()],
    plots=[
        Histogram(x="score_fluency"),
        Histogram(x="score_coherence"),
        Histogram(x="score_politeness"),
        Histogram(x="score_grammar"),
    ],
)

CheckToneQuality = lambda llm_persona: Check(
    name="tone_critique_score",
    operators=[ToneCritique(llm_persona=llm_persona)],
    plots=[Histogram(x="score_tone")],
)

# -----------------------------------------------------------
# Guideline related
# -----------------------------------------------------------

CheckGuidelineAdherence = lambda guideline, guideline_name="guideline", response_schema=None: Check(
    name = f"{guideline_name}_adherence_score",
    operators=[GuidelineAdherenceScore(guideline=guideline, guideline_name=guideline_name, response_schema=response_schema)],
    plots=[Histogram(x=f"score_{guideline_name}_adherence")]
)

# -----------------------------------------------------------
# Prompt Injection related
# -----------------------------------------------------------

CheckPromptInjection = lambda: Check(
    name = "prompt_injection_score",
    operators=[PromptInjectionScore()],
    plots=[Histogram(x="score_prompt_injection")]
)

# -----------------------------------------------------------
# Conversation related
# -----------------------------------------------------------

CheckConversationSatisfaction = lambda user_persona = "user", llm_persona = None : Check(
    name = "conversation_satisfaction_score",
    operators=[ConversationSatisfactionScore(user_persona, llm_persona)],
    plots=[Histogram(x=f"score_conversation_satisfaction")]
)


# -----------------------------------------------------------
# Response Matching related
# -----------------------------------------------------------

CheckResponseMatching = lambda method = "llm": Check(
    name = f"{method}_score",
    operators=[ResponseMatchingScore(method=method)],
    plots=[Histogram(x=f"{method}_score")]
)



from uptrain.v0.core.classes.measurables import (
    Measurable,
    InputFeatureMeasurable,
    OutputFeatureMeasurable,
    FeatureMeasurable,
    FeatureConcatMeasurable,
    ConditionMeasurable,
    CustomMeasurable,
    AccuracyMeasurable,
    MAEMeasurable,
    MAPEMeasurable,
    ScalarFromEmbeddingMeasurable,
    DistanceMeasurable,
    RecHitRateMeasurable,
)
from uptrain.v0.ee.classes.measurables import (
    GrammerScoreMeasurable,
)

from uptrain.v0.constants import MeasurableType


class MeasurableResolver:
    """Class that resolves a measurable key to a measurable class instance."""

    def __init__(self, args) -> None:
        super().__init__()
        self._args = args

    def has_valid_resolve_args(self) -> bool:
        if self._args is None:
            return False
        if len(self._args) == 0:
            return False
        return True

    def resolve(self, framework) -> Measurable:
        if not self.has_valid_resolve_args():
            return None

        resolve_args = self._args
        measurable_type = resolve_args["type"]
        if measurable_type == MeasurableType.INPUT_FEATURE:
            return InputFeatureMeasurable(framework, resolve_args["feature_name"])
        elif measurable_type == MeasurableType.FEATURE_CONCAT:
            return FeatureConcatMeasurable(framework, resolve_args["feat_name_list"])
        elif measurable_type == MeasurableType.PREDICTION:
            return OutputFeatureMeasurable(framework)
        elif measurable_type == MeasurableType.CUSTOM:
            return CustomMeasurable(framework, resolve_args)
        elif measurable_type == MeasurableType.ACCURACY:
            return AccuracyMeasurable(framework)
        elif measurable_type == MeasurableType.MAE:
            return MAEMeasurable(framework)
        elif measurable_type == MeasurableType.MAPE:
            return MAPEMeasurable(framework)
        elif measurable_type == MeasurableType.CONDITION_ON_INPUT:
            return ConditionMeasurable(
                framework,
                InputFeatureMeasurable(framework, resolve_args["feature_name"]),
                resolve_args["condition_args"],
            )
        elif measurable_type == MeasurableType.CONDITION_ON_PREDICTION:
            return ConditionMeasurable(
                framework,
                OutputFeatureMeasurable(framework),
                resolve_args["condition_args"],
            )
        elif measurable_type == MeasurableType.SCALAR_FROM_EMBEDDING:
            return ScalarFromEmbeddingMeasurable(
                framework, resolve_args["idx"], resolve_args["extract_from"]
            )
        elif measurable_type == MeasurableType.DISTANCE:
            return DistanceMeasurable(
                framework,
                resolve_args["base"],
                resolve_args["reference"],
                resolve_args["distance_types"],
            )
        elif measurable_type == MeasurableType.REC_HIT_RATE:
            return RecHitRateMeasurable(framework)
        elif measurable_type == MeasurableType.GRAMMAR_SCORE:
            return GrammerScoreMeasurable(
                framework, resolve_args.get("feature_name", None)
            )
        else:
            raise Exception("Resolver not defined")



import os

try:
    import psycopg2
except:
    psycopg2 = None

from uptrain.v0.core.lib.helper_funcs import dependency_required


@dependency_required(psycopg2, "psycopg2")
class PostgresLogs:
    def __init__(self, db_name):
        self.conn = psycopg2.connect(
            database=db_name, user=os.environ['POSTGRES_USERNAME'], password=os.environ['POSTGRES_PASSWORD'])
        self.conn.autocommit = True
        self.cursor = self.conn.cursor()
        self.tables_created = {}


    def add_scalars(self, dict, table_name, update_val=False):
        if self.tables_created.get(table_name, 0) == 0:
            schema = "(uptrain_log_id serial primary key, "
            for key, val in dict.items():
                if isinstance(val, int):
                    type = "INTEGER"
                elif isinstance(val, str):
                    type = "VARCHAR"
                elif isinstance(val, float):
                    type = "float"
                else:
                    raise Exception(f'{type(val)} not mapped to postgres')
                schema += key + " " + type + " NOT NULL, "
            schema = schema[0:-2] + ")"
            self.cursor.execute(f"""CREATE TABLE {table_name} {schema}""")
            print(f"Created table: {table_name}")
            self.tables_created[table_name] = 1

        if update_val:
            raise Exception("Not supported")
        else:
            keys = ', '.join(list(dict.keys()))
            vals = []
            for val in list(dict.values()):
                if isinstance(val, str):
                    vals.append(f"'{val}'")
                else:
                    vals.append(str(val))
            vals = ', '.join(vals)
            self.cursor.execute(f"""INSERT INTO {table_name} ({keys}) VALUES ({vals});""")
            # self.cursor.execute(f"SELECT * FROM {table_name}")
            # rows = self.cursor.fetchall()
            # for row in rows:
            #     print(row)



import os
import numpy as np
from torch import nn
import torch
from . import KpsDataset, read_json

import random

random.seed(42)
torch.manual_seed(42)


class BinaryClassification(nn.Module):
    def __init__(self):
        super(BinaryClassification, self).__init__()
        # Number of input features is 34.
        self.layer_1 = nn.Linear(34, 64)
        self.layer_out = nn.Linear(64, 1)
        self.relu = nn.ReLU()
        self.dropout = nn.Dropout(p=0.1)

    def forward(self, inputs):
        x = self.relu(self.layer_1(inputs))
        x = self.dropout(x)
        x = self.layer_out(x)
        return x


def get_accuracy_torch(
    testing_file, model_save_name, model_dir="trained_models_torch/"
):
    testing_dataset = KpsDataset(
        testing_file,
        batch_size=len(read_json(testing_file)),
        shuffle=False,
        augmentations=False,
        is_test=True,
    )
    model = BinaryClassification()
    model.load_state_dict(torch.load(model_dir + model_save_name))
    model.eval()
    print(
        "Evaluating model:",
        model_save_name,
        " on ",
        len(read_json(testing_file)),
        " data-points",
    )
    with torch.inference_mode():
        pred_classes = []
        gt_classes = []

        for ele in testing_dataset:
            X_test = ele[0]["kps"]
            y_test = ele[1]

            X_test, y_test = torch.tensor(X_test).type(torch.float), torch.tensor(
                y_test
            ).type(torch.float)
            test_logits = model(X_test).squeeze()
            pred_class = torch.round(torch.sigmoid(test_logits))
            pred_classes.extend(pred_class)
            gt_classes.extend(list(ele[1]))
    accuracy = np.sum(np.array(gt_classes) == np.array(pred_classes)) / len(
        pred_classes
    )
    return accuracy


def train_model_torch(
    training_file, model_save_name, model_dir="trained_models_torch/"
):
    model_loc = model_dir + model_save_name
    os.makedirs(model_dir, exist_ok=True)
    if os.path.exists(model_loc):
        return
    print(
        "Training on: ",
        training_file,
        " which has ",
        len(read_json(training_file)),
        " data-points",
    )
    training_dataset = KpsDataset(training_file, shuffle=True, augmentations=True)
    model = BinaryClassification()
    loss_fn = nn.BCEWithLogitsLoss()
    optimizer = torch.optim.Adam(params=model.parameters(), lr=5e-4)
    torch.manual_seed(22)
    epochs = 2
    for epoch in range(epochs):
        model.train()
        loss_sum = 0
        for X_train, y_train in training_dataset:
            X_train = torch.tensor(X_train).type(torch.float)
            y_train = torch.tensor(y_train).type(torch.float)
            y_logits = model(X_train).squeeze()
            loss = loss_fn(y_logits, y_train)
            optimizer.zero_grad()
            loss.backward()
            optimizer.step()
            loss_sum += loss.item()
        print(f"Epoch {epoch}: Loss {loss_sum/len(training_dataset)}")

    torch.save(model.state_dict(), model_loc)
    print("Model saved at: ", model_loc)



import os
import sys
import json
import random

import streamlit as st
import pandas as pd
import plotly.graph_objects as go
import numpy as np
import plotly.express as px

from uptrain.v0.core.lib.helper_funcs import make_dir_friendly_name

# -----------------------------------------------------------
# Set up layout of the dashboard and the sidebar
# -----------------------------------------------------------

st.set_page_config(
    page_title="UpTrain Dashboard",
    layout="wide",
    page_icon="https://github.com/uptrain-ai/uptrain/raw/dashboard/uptrain/core/classes/logging/uptrain_logo_icon.png",
)
st.title("UpTrain Live Dashboard")
st_style = """<style> footer {visibility: hidden;} </style>"""
st.markdown(st_style, unsafe_allow_html=True)


# -----------------------------------------------------------
# Read parameters for this run
# -----------------------------------------------------------


def read_config(log_folder: str):
    """read the uptrain configuration for this run"""
    config_file = os.path.join(log_folder, "config.json")
    with open(config_file, encoding="utf-8") as f:
        return json.loads(f.read())


INPUT_LOG_FOLDER = sys.argv[1]
LOG_FOLDER = INPUT_LOG_FOLDER  # for partitioned datasets, this could be a subdirectory of INPUT_LOG_FOLDER

if "config.json" in os.listdir(INPUT_LOG_FOLDER):
    CONFIG = read_config(INPUT_LOG_FOLDER)
else:
    print(
        f"Could not find an Uptrain config at {INPUT_LOG_FOLDER}. Looking into sub-dirs."
    )
    all_sub_dirs = [
        os.path.join(INPUT_LOG_FOLDER, d) for d in os.listdir(INPUT_LOG_FOLDER)
    ]
    valid_sub_dirs = [
        d for d in all_sub_dirs if os.path.isdir(d) and "config.json" in os.listdir(d)
    ]
    if not len(valid_sub_dirs):
        st.warning("No uptrain config found. Please check the log folder.")
        st.stop()
    else:
        with st.sidebar:
            st.subheader("Select the dataset partition to analyze")
            LOG_FOLDER = st.selectbox(
                "Partition",
                [""] + valid_sub_dirs,
                format_func=lambda x: x.split("/")[-1],
            )
        if LOG_FOLDER == "":
            st.stop()
        else:
            assert LOG_FOLDER is not None
            CONFIG = read_config(LOG_FOLDER)

# verify there are checks in the config
if not len(CONFIG["checks"]):
    st.warning("No checks found in the logs configured.")
    st.stop()

# -----------------------------------------------------------
# Plotting routines
# -----------------------------------------------------------


def get_plotname_n_file_for_statistic(
    check: dict, distance_type: str
) -> tuple[str, str]:
    """get the path to the log file for the dashboard.

    TODO: this logic should be defined on the actual Check objects. This only works for Statistics.
    """
    dashboard_name = make_dir_friendly_name(check["type"])
    plot_name = make_dir_friendly_name(
        distance_type
        if "reference" not in check
        else "_".join([distance_type, check["reference"]])
    )

    return (
        f"{dashboard_name} - {plot_name}",
        os.path.join(LOG_FOLDER, dashboard_name, f"{plot_name}.csv"),
    )


def apply_model_n_feature_filters(
    df: pd.DataFrame, model_variant: dict, feature_filters: dict
) -> pd.DataFrame:
    list_conds = []
    for feature, value in feature_filters.items():
        list_conds.append(df[feature] == value)
    for k, v in model_variant.items():
        list_conds.append(df[k] == v)
    return df[np.logical_and.reduce(list_conds)]


def plot_check_distance(check: dict, model_variant: dict, feature_filters: dict):
    x_log = st.checkbox(
        "log x", help="See x-axes in log-scale", key=check["type"] + "log_x"
    )
    for dist_type in check["distance_types"]:
        plot_title, fname = get_plotname_n_file_for_statistic(check, dist_type)
        st.markdown(f"### Line chart for {plot_title}")

        if os.path.getsize(fname) == 0:
            st.warning("No data found for the specified check.")
            return
        else:
            df = pd.read_csv(fname)

        df = apply_model_n_feature_filters(df, model_variant, feature_filters)
        if len(df) == 0:
            st.warning("No data found for the specified filters.")
            return

        grouping_col = check["aggregate_args"]["feature_name"]
        xaxis = check["count_args"]["feature_name"]
        yaxis = "check"

        if df[grouping_col].nunique() > 1000:
            # pick 1000 random values
            unique_values = df[grouping_col].unique()
            random.seed(42)
            random.shuffle(unique_values)
            df = df[df[grouping_col].isin(unique_values[:1000])]

        df[grouping_col] = df[grouping_col].astype(str)
        fig = px.line(df, x=xaxis, y=yaxis, color=grouping_col, log_x=x_log)
        st.plotly_chart(fig, use_container_width=True)


def plot_check_convergence(check: dict, model_variant: dict, feature_filters: dict):
    for dist_type in check["distance_types"]:
        plot_title, fname = get_plotname_n_file_for_statistic(check, dist_type)
        st.markdown(f"### Histogram for {plot_title}")

        if os.path.getsize(fname) == 0:
            st.warning("No data found for the specified check.")
            return
        else:
            df = pd.read_csv(fname)

        df = apply_model_n_feature_filters(df, model_variant, feature_filters)
        if len(df) == 0:
            st.warning("No data found for the specified filters.")
            return

        grouping_col = check["count_args"]["feature_name"]
        value_col = "check"
        df[grouping_col] = df[grouping_col].astype(str)
        fig = px.histogram(df, x=value_col, color=grouping_col)
        st.plotly_chart(fig, use_container_width=True)


def plot_visual_umap(check: dict, model_variant: dict, feature_filters: dict):
    dashboard_name = make_dir_friendly_name(check.get("dashboard_name", "visual"))
    plot_name = make_dir_friendly_name(check["type"])
    fname = os.path.join(sys.argv[1], dashboard_name, f"{plot_name}.json")

    plot_title = f"{dashboard_name} - {plot_name}"
    st.markdown(f"### {plot_title}")
    with open(fname, encoding="utf-8") as f:
        all_data = [json.loads(line) for line in f.readlines()]

    data = None
    for row in all_data:
        valid_model = all(
            row[model_arg] == model_variant[model_arg] for model_arg in model_variant
        )
        valid_features = all(
            row[feature_arg] == feature_filters[feature_arg]
            for feature_arg in feature_filters
        )
        if valid_model and valid_features:
            data = row
            break

    if data is None:
        st.warning("No data found for the specified filters.")
        st.stop()

    arr = np.array(data["umap"])
    labels_dict = data["labels"]
    color_type = st.selectbox("Select coloring type", labels_dict.keys())
    x = arr[:, 0]
    y = arr[:, 1]
    dictn = {"x": x, "y": y, "color": np.array(labels_dict[color_type])}
    if arr.shape[1] == 3:
        dictn.update({"z": arr[:, 2]})

    hover_data = []
    if "hover_texts" in data:
        dictn.update({"hover": data["hover_texts"]})
        if len(data["hover_texts"]):
            hover_data = list(data["hover_texts"][0].keys())

    for key in data.keys():
        if key not in ["umap", "labels", "hover_texts"]:
            dictn.update({key: data[key]})
    df = pd.DataFrame(dictn)

    if "hover" in list(df.columns):
        hover_df = pd.DataFrame(list(df["hover"]))
        for key in hover_data:
            if key not in hover_df.columns:
                df[key] = [""] * len(hover_df)
            else:
                df[key] = pd.Series(hover_df[key]).fillna("").tolist()
    else:
        hover_df = pd.DataFrame()

    if arr.shape[1] == 2:
        fig = px.scatter(df, x="x", y="y", color="color", hover_data=hover_data)
    elif arr.shape[1] == 3:
        z = list(df["z"])
        fig = px.scatter_3d(
            df, x="x", y="y", z="z", color="color", hover_data=hover_data
        )
    else:
        raise Exception("Umap dimension not 2D or 3D")
    st.plotly_chart(fig, use_container_width=True)


def plot_dashboard(check: dict, model_variant: dict, feature_filters: dict):
    if check["type"] == "distance":
        plot_check_distance(check, model_variant, feature_filters)
    elif check["type"] == "convergence_stats":
        plot_check_convergence(check, model_variant, feature_filters)
    elif check["type"] in ("UMAP", "t-SNE"):
        plot_visual_umap(check, model_variant, feature_filters)
    else:
        st.warning(f"Checks of type: {check['type']} are not supported yet.")


# -----------------------------------------------------------
# Create a sidebar for the user to specify what they wanna see
# -----------------------------------------------------------

with st.sidebar:
    st.subheader("Select the dashboard to view")
    selected_check = st.radio(
        "Checks",
        options=CONFIG["checks"],
        format_func=lambda x: str(x["type"]),
    )
    st.markdown("---")
    assert selected_check is not None

    # Work with a specific slice of the data
    feature_filters = {}
    if "feature_args" in selected_check:
        st.subheader("Pick a specific slice of the logged data")
        for i, feature in enumerate(selected_check["feature_args"]):
            feature_name = feature["feature_name"]
            selected_value = st.sidebar.selectbox(
                feature_name, ["All", *feature["allowed_values"]]
            )
            if selected_value != "All":
                feature_filters.update({feature_name: selected_value})

    # select a model variant from the cartesian product of all model args
    model_variant = {}
    if "model_args" in selected_check:
        st.subheader("Select model variant")
        for i, model in enumerate(selected_check["model_args"]):
            model_name = model["feature_name"]
            selected_value = st.sidebar.selectbox(model_name, model["allowed_values"])
            model_variant.update({model_name: selected_value})


plot_dashboard(selected_check, model_variant, feature_filters)



"""Basic IO operators for reading and writing data from Uptrain."""

from __future__ import annotations
import typing as t

import polars as pl
import pandas as pd
if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *
from uptrain.utilities import lazy_load_dep

# -----------------------------------------------------------
# Read from text file formats - csv, json
# -----------------------------------------------------------


@register_op
class CsvReader(TransformOp):
    """Reads data from a csv file.

    Attributes:
        fpath (str): Path to the csv file.
        batch_size (Optional[int]): Number of rows to read at a time. Defaults to None, which reads the entire file.

    """

    fpath: str
    batch_size: t.Optional[int] = None

    def setup(self, settings: Settings):
        self._executor = TextReaderExecutor(self)
        return self

    def run(self) -> TYPE_TABLE_OUTPUT:
        return {"output": self._executor.run()}


@register_op
class JsonReader(TransformOp):
    """Reads data from a json file.

    Attributes:
        fpath (str): Path to the json file.
        batch_size (Optional[int]): Number of rows to read at a time. Defaults to None, which reads the entire file.

    """

    fpath: str
    batch_size: t.Optional[int] = None

    def setup(self, settings: Settings):
        self._executor = TextReaderExecutor(self)
        return self

    def run(self) -> TYPE_TABLE_OUTPUT:
        return {"output": self._executor.run()}


class TextReaderExecutor:
    op: t.Union[CsvReader, JsonReader]
    dataset: pl.DataFrame
    rows_read: int

    def __init__(self, op: t.Union[CsvReader, JsonReader]):
        self.op = op
        self.rows_read = 0
        if isinstance(self.op, CsvReader):
            self.dataset = pl.read_csv(self.op.fpath)
        elif isinstance(self.op, JsonReader):
            self.dataset = pl.read_ndjson(self.op.fpath)
            null_count = self.dataset.null_count().to_dicts()[0]
            for _, value in null_count.items():
                if value == self.dataset.shape[0]:
                    ## read from pandas
                    pd_df = pd.read_json(self.op.fpath, lines=True)
                    self.dataset = pl.DataFrame(pd_df)
                    break

    @property
    def is_incremental(self) -> bool:
        return self.op.batch_size is not None

    def run(self) -> pl.DataFrame | None:
        if not self.is_incremental:
            data = self.dataset
        else:
            if self.rows_read >= len(self.dataset):
                data = None
            else:
                assert self.op.batch_size is not None
                self.rows_read += self.op.batch_size
                data = self.dataset.slice(self.rows_read, self.op.batch_size)
        return data


# -----------------------------------------------------------
# Read from a Delta Table, which uses parquet files under the hood
# -----------------------------------------------------------


@register_op
class DeltaReader(TransformOp):
    """Reads data from a Delta Lake table.

    Attributes:
        fpath (str): File path to the Delta Lake table.
        batch_split (bool): Whether to read the table in batches. Defaults to False.

    """

    fpath: str
    batch_split: bool = False
    _dataset: t.Any  # pyarrow dataset
    _batch_generator: t.Optional[t.Iterator[t.Any]]  # record batch generator

    def setup(self, settings: Settings):

        lazy_load_dep("pyarrow", "pyarrow>=10.0.0")
        dl = lazy_load_dep("deltatable", "deltalake>=0.9")

        self._dataset = dl.DeltaTable(self.fpath).to_pyarrow_dataset()
        if self.is_incremental:
            self._batch_generator = iter(self._dataset.to_batches())
        return self

    @property
    def is_incremental(self) -> bool:
        return self.batch_split is True

    def run(self) -> TYPE_TABLE_OUTPUT:
        if not self.is_incremental:
            data = pl.from_arrow(self._dataset.to_table())
        else:
            try:
                data = pl.from_arrow(next(self._batch_generator))  # type: ignore
            except StopIteration:
                data = None

        if data is not None:
            assert isinstance(data, pl.DataFrame)
        return {"output": data}


# -----------------------------------------------------------
# Writer objects
# -----------------------------------------------------------


@register_op
class DeltaWriter(OpBaseModel):
    fpath: str
    columns: t.Optional[list[str]] = None

    def setup(self, settings: Settings):
        dl = lazy_load_dep("deltatable", "deltalake>=0.9")

        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        if self.columns is None:
            self.columns = list(data.columns)
        assert set(self.columns) == set(data.columns)
        data.write_delta(self.fpath, mode="append")
        return {"output": data}

    def to_reader(self):
        return DeltaReader(fpath=self.fpath)  # type: ignore


@register_op
class JsonWriter(OpBaseModel):
    fpath: str
    columns: t.Optional[list[str]] = None

    def setup(self, settings: Settings):
        return self

    def to_reader(self):
        return JsonReader(fpath=self.fpath)  # type: ignore

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        if self.columns is None:
            self.columns = list(data.columns)
        assert set(self.columns) == set(data.columns)
        with open(self.fpath, "a") as f:
            f.write(data.write_ndjson())
        return {"output": data}



import numpy as np

from typing import Any, Dict

from uptrain.v0.core.classes.measurables import Measurable
from uptrain.v0.core.classes.signals import SignalManager


class CustomMeasurable(Measurable):
    """Class that computes a custom metric measurable using a user-defined signal formula."""

    def __init__(self, framework, custom_args: Dict[str, Any] = {}) -> None:
        """Constructor for CustomMeasurable class.

        Parameters
        ----------
        framework
            UpTrain framework object
        custom_args
            A dictionary containing custom arguments for usage in computation.
            The dictionary must contain "signal_formulae" as key which maps to
            a callback that takes in inputs, outputs, gts and extra_args, and
            returns a boolean value as the signal measure.
        """
        super().__init__(framework)
        self._args = custom_args
        self.signal_manager = SignalManager()
        self.signal_manager.add_signal_formulae(self._args["signal_formulae"])

    def _compute(self, inputs=None, outputs=None, gts=None, extra=None) -> np.ndarray:
        val = self.signal_manager.evaluate_signal(
            inputs, outputs, gts=gts, extra_args=extra
        )
        return val

    def col_name(self) -> str:
        return str(self.signal_manager)

    # TODO: Decommission and find a generic way
    def extract_val_from_training_data(self, x):
        fake_inputs = {}
        for key, val in x.items():
            fake_inputs.update({key: np.array([val])})
        return self.signal_manager.evaluate_signal(fake_inputs, None)



"""
Implement oeprators over sql data.
"""

from __future__ import annotations

import itertools
import json
import os
import typing as t

from pydantic import BaseModel
import polars as pl

from uptrain.utilities import lazy_load_dep
from uptrain.utilities.sql_utils import (
    extract_tables_and_columns,
    extract_tables_and_columns_from_create,
    PLACEHOLDER_TABLE,
    execute_and_compare_sql,
)

if t.TYPE_CHECKING:
    from uptrain.framework.base import *
from uptrain.operators.base import *

sqlglot = lazy_load_dep("sqlglot", "sqlglot")


# TODO: define a Table object and dump that into the dataframe
class Table(BaseModel):
    create_table_stmt: str
    table_name: str
    columns: list


@register_op
class ParseCreateStatements(TransformOp):
    """
    Read tables and columns from ";" separated CREATE TABLE statements and writes a json dictionary Table -> [columns].

    Attributes:
        col_in_schema_def (str): Column name of schema def containing CREATE TABLE statements.
        col_out_tables (str): Column to write parsed tables and columns.

    """

    col_in_schema_def: str
    col_out_tables: str

    def setup(self, settings: Settings):
        return self

    def __fetch_tables_columns(self, create_statements) -> t.Tuple(str, str, str):
        tables_and_columns = {}
        # SQL statements are separated by ';'
        statements = create_statements.split(";")

        # Create table statements
        for statement in statements:
            if statement.upper().strip().startswith("CREATE TABLE"):
                # Add the statement to the list
                statement = statement.strip()
                # TODO: handle input database dialect instead of assuming it.
                parsed = sqlglot.parse(statement, read=sqlglot.Dialects.SQLITE)
                table, columns = extract_tables_and_columns_from_create(parsed[0])
                tables_and_columns[table] = list(columns)

        return json.dumps(tables_and_columns)

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        schemas = data.get_column(self.col_in_schema_def)
        tables = [self.__fetch_tables_columns(schema) for schema in schemas]
        return {"output": data.with_columns([pl.Series(self.col_out_tables, tables)])}


@register_op
class ParseSQL(TransformOp):
    """
    Read tables and columns from a generic SQL SELECT statement and writes a json dictionary Table -> [columns].
    Note that we don't use table schema definition to do this but instead simply parse the SQL. Output might have a
    placeholder table to include columns that are accessed without a table descriptor.

    This is typically used along with ValidateTables to validate tables and columns in the predicted SQL.

    Attributes:
        col_in_sql (str): Column of input SQL containing SQL SELECT statement.
        col_out_tables (str): Column to write parsed tables and columns.
        col_out_is_valid_sql (str): Column to store if sql is valid as per sql parser.

    """

    col_in_sql: str
    col_out_tables: str
    col_out_is_valid_sql: str

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        sqls = data.get_column(self.col_in_sql)
        tables = []
        is_valid = []
        for sql in sqls:
            try:
                # TODO: parse using expected dialect
                parsed = sqlglot.parse(sql)
                tables_and_columns = extract_tables_and_columns(parsed[0])
                # Since sets are not serializable, convert to list
                for table, columns in tables_and_columns.items():
                    tables_and_columns[table] = list(columns)
                tables.append(json.dumps(tables_and_columns))
                is_valid.append(True)
            except sqlglot.errors.ParseError:
                tables.append(json.dumps({}))
                is_valid.append(False)

        return {
            "output": data.with_columns(
                [
                    pl.Series(self.col_out_tables, tables),
                    pl.Series(self.col_out_is_valid_sql, is_valid),
                ]
            )
        }


@register_op
class ValidateTables(TransformOp):
    """
    Ensures that table and column names from response/predicted SQL are valid tables columns as per the schema
    definition.

    This is typically used with ParseSQL and ParseCreateStatements.

    Attributes:
        col_in_response_tables (str): Column containing response SQL tables and columns as a json dict Table -> [columns].
        col_in_schema_tables (str): Column containing schema definition tables and columns as a json dict Table -> [columns].
        col_out_is_tables_valid (str): Column to store if tables are valid.
        col_out_is_cols_valid (str): Column to store if columns are valid.

    """

    col_in_response_tables: str
    col_in_schema_tables: str
    col_out_is_tables_valid: str
    col_out_is_cols_valid: str

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        response_tables = data.get_column(self.col_in_response_tables)
        schema_tables = data.get_column(self.col_in_schema_tables)
        results = []
        results_column = []
        for response_table, schema_table in zip(response_tables, schema_tables):
            is_valid_table = True
            is_valid_column = True
            # deserialize json
            s = json.loads(schema_table)
            r = json.loads(response_table)
            columns_list = [columns for table, columns in s.items()]
            all_columns = set(itertools.chain(*columns_list))
            for table, columns in r.items():
                is_valid_table &= table == PLACEHOLDER_TABLE or table in s
                # TODO: handle aliases and do lineage check
                is_valid_column = (
                    is_valid_column
                    and is_valid_table
                    and (
                        (
                            table != PLACEHOLDER_TABLE
                            and set(columns).issubset(set(s[table]))
                        )
                        or (
                            table == PLACEHOLDER_TABLE
                            and set(columns).issubset(all_columns)
                        )
                    )
                )

            results.append(is_valid_table)
            results_column.append(is_valid_column)
        return {
            "output": data.with_columns(
                [
                    pl.Series(self.col_out_is_tables_valid, results),
                    pl.Series(self.col_out_is_cols_valid, results_column),
                ]
            )
        }


@register_op
class ExecuteAndCompareSQL(TransformOp):
    """
    Execute predicted SQL, ground truth SQL and compute execution accuracy of the predicted sql.

    For now, we expect the output to exactly match along with the column names.

    ignore_column_order, ignore_row_order params attempt to do a semantic match by ignoring the order. This allows us to
    correctly compare `SELECT a,b` and `SELECT b,a` if the column order is not important. However, the intent in the
    text query also needs to be taken in consideration for correct sematic match.

    Attributes:
        col_in_response_sql (str): Column containing response SQL.
        col_in_gt_sql (str): Column containing schema definition tables and columns as a json dict Table -> [columns].
        col_in_db_path (str): Column to store if tables are valid.
        col_out_execution_accuracy (str): Column to store if columns are valid.
        ignore_column_order (bool): Boolean param to ignore column order when comparing SQL output. True by default.
        ignore_row_order (bool): Boolean param to ignore row order when comparing SQL output. True by default.

    """

    col_in_response_sql: str
    col_in_gt_sql: str
    col_in_db_path: str
    col_out_execution_accuracy: str
    ignore_column_order: bool = True
    ignore_row_order: bool = True

    def setup(self, settings: Settings):
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        response_sqls = data.get_column(self.col_in_response_sql)
        gt_sqls = data.get_column(self.col_in_gt_sql)
        db_paths = data.get_column(self.col_in_db_path)
        results = []
        for response_sql, gt_sql, db_path in zip(response_sqls, gt_sqls, db_paths):
            results.append(
                execute_and_compare_sql(
                    response_sql,
                    gt_sql,
                    db_path,
                    ignore_column_order=self.ignore_column_order,
                    ignore_row_order=self.ignore_row_order,
                )
            )
        return {
            "output": data.with_columns(
                [pl.Series(self.col_out_execution_accuracy, results)]
            )
        }



"""
Implement operators to evaluate conversation datapoints for a 
LLM-powered agent pipeline. 
"""

from __future__ import annotations
import typing as t

from loguru import logger
import polars as pl

if t.TYPE_CHECKING:
    from uptrain.framework import Settings
from uptrain.operators.base import *


@register_op
class ConversationSatisfactionScore(ColumnOp):
    """
    Grade if the user is satifised with the conversation with the ChatBot. 

     Attributes:
        col_conversation (str): Column name for the stored conversations
        user_perosna (str): The persona of user asking the queries.
        llm_persona (str): The persona of LLM responding to the queries.
    Raises:
        Exception: Raises exception for any failed evaluation attempts

    """
    col_conversation: str = "conversation"
    col_out: str = "score_conversation_satisfaction"
    llm_persona: t.Union[str, None] = None
    user_persona: str = 'user'

    def setup(self, settings: t.Optional[Settings] = None):
        from uptrain.framework.remote import APIClient

        assert settings is not None
        self._api_client = APIClient(settings)
        return self

    def run(self, data: pl.DataFrame) -> TYPE_TABLE_OUTPUT:
        data_send = [
            {
                "conversation": row[self.col_conversation]
            }
            for row in data.to_dicts()
        ]
        try:
            results = self._api_client.evaluate(
                "ConversationSatisfaction", data_send, {
                    "user_persona": self.user_persona,
                    "llm_persona": self.llm_persona
                })
    
        except Exception as e:
            logger.error(f"Failed to run evaluation for `ConversationSatisfactionScore`: {e}")
            raise e

        assert results is not None
        return {"output": data.with_columns(pl.from_dicts(results).rename({"score_conversation_satisfaction": self.col_out}))}



__all__ = [
    "builtins",
    "Check",
    "CheckSet",
    "ExperimentArgs",
    "EvalLLM",
    "OperatorDAG",
    "Settings",
    "Signal",
    "APIClient",
    "DataSchema",
    "Evals",
    "CritiqueTone",
    "GuidelineAdherence",
    "ResponseMatching",
    "ConversationSatisfaction"
]

from .base import OperatorDAG, Settings
from .checks import CheckSet, Check, ExperimentArgs
from .signal import Signal
from .remote import APIClient, DataSchema
from .builtins import builtins
from .evals import Evals, CritiqueTone, GuidelineAdherence, ResponseMatching, ConversationSatisfaction
from .evalllm import EvalLLM



import numpy as np
import scipy
import pandas as pd

from uptrain.v0.core.classes.monitors import AbstractMonitor
from uptrain.v0.constants import Monitor
from uptrain.v0.core.lib.helper_funcs import read_json
from uptrain.v0.core.classes.algorithms import Clustering


class FeatureDrift(AbstractMonitor):
    dashboard_name = "feature_drift"

    def model_choices(self, check):
        return [{"is_embedding": False}]

    def base_init(self, fw, check):
        self.reference_dataset = check["reference_dataset"]
        self.NUM_BUCKETS = check.get("num_buckets", 10)
        self.psi_threshold = check.get("psi_threshold", 0.2)
        self.step = check.get("step", 2000)
        self.prod_dist_counts_arr = []
        self.all_count = 0
        self.plot_name = "Feature " + self.measurable.col_name()
        self.feats = np.array([])
        self.psis = []

        self.train_dist, self.train_bins = self.get_ref_data_stats()

        # convert the array self.train_bins into tuples of (min, max) values
        self.bar_vals = []
        for i in range(len(self.train_bins) - 1):
            start_str = "{:.2f}".format(self.train_bins[i])
            end_str = "{:.2f}".format(self.train_bins[i + 1])
            self.bar_vals.append(start_str + "-" + end_str)

    def check(self, inputs, outputs, gts=None, extra_args={}):
        """
        Find the population stability index between the reference and
        production distribtions by assigning them into
        reference buckets.
        """

        feats = self.measurable.compute_and_log(
            inputs=inputs, outputs=outputs, gts=gts, extra=extra_args
        )
        self.all_count += len(feats)
        self.feats = np.append(self.feats, np.squeeze(np.array(feats)))

        if len(self.feats) >= self.step:
            prod_dist, _ = np.histogram(self.feats, bins=self.train_bins)
            prod_dist = prod_dist / sum(prod_dist)

            # Find the PSI between the two distributions
            psi = self.get_psi(self.train_dist, prod_dist)
            self.psis.append(psi)

            self.log_handler.add_bar_graphs(
                self.plot_name,
                {
                    "reference": dict(zip(self.bar_vals, list(self.train_dist))),
                    "production": dict(zip(self.bar_vals, list(prod_dist))),
                },
                self.dashboard_name,
            )

            self.log_handler.add_scalars(
                "psi",
                {"y_psi": psi},
                self.all_count,
                self.dashboard_name,
                file_name=self.measurable.col_name(),
            )
            feat_name = self.measurable.feature_name

            # High PSI alerts the user to a feature drift
            if psi > self.psi_threshold:
                alert = f"Feature Drift last detected at {self.all_count} for {feat_name} with PSI = {psi}"
                self.log_handler.add_alert(
                    f"Feature Drift Alert for {feat_name} 🚨", alert, self.dashboard_name
                )
            self.feats = np.array([])

    def get_ref_data_stats(self):
        """
        Find the bucketing scheme for data in ref_arr. The data can be numerical or categorical.
        """
        if self.reference_dataset.split(".")[-1] == "json":
            data = read_json(self.reference_dataset)
            all_inputs = np.array(
                [self.measurable.extract_val_from_training_data(x) for x in data]
            )
        elif self.reference_dataset.split(".")[-1] == "csv":
            data = pd.read_csv(self.reference_dataset).to_dict()
            for key in data:
                data[key] = list(data[key].values())
            all_inputs = np.array(self.measurable.extract_val_from_training_data(data))
        else:
            raise Exception("Reference data file type not recognized.")

        prob = np.linspace(0, 1, self.NUM_BUCKETS + 1)
        bins = scipy.stats.mstats.mquantiles(all_inputs, prob)
        vals, bins = np.histogram(all_inputs, bins=bins)
        return vals / sum(vals), bins

    def get_psi(self, ref_dist, prod_dist):
        """
        Find the population stability index between the two distributions.
        """
        return np.sum(
            np.where(
                ref_dist == 0, 0, (ref_dist - prod_dist) * np.log(ref_dist / prod_dist)
            )
        )
